/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h> 
#include <utime.h>

/* compare two local files, 1 if different */ 
static int diff(const char *x, const char *y) { 
  FILE *xfd, *yfd; 
  if (!(xfd=fopen(x,"r"))) return 1; 
  if (!(yfd=fopen(y,"r"))) { fclose(xfd); return 1; } 
  /* yes, I know this is dumb */ 
  while (! feof(xfd) && ! feof(yfd)) { 
    if (getc(xfd) != getc(yfd)) { fclose(xfd); fclose(yfd); return 1; } 
  } 
  return 0; 
}

/* copy a file if necessary, not trashing the original unless necessary */ 
foreign_t 
pl_copy(term_t sour, term_t dest) { 
  char *sour_val; 
  char *dest_val; 
  FILE *sfd, *dfd; 
  struct stat sbuf, dbuf; 
  struct utimbuf times; 
  if (!PL_get_atom_chars(sour,&sour_val)) PL_fail; 
/* fprintf(stderr, "got sour_val\n");  */
  if (!PL_get_atom_chars(dest,&dest_val)) PL_fail; 
/* fprintf(stderr, "got dest_val\n");  */
  if (stat(sour_val,&sbuf)!=0) PL_fail;
/* fprintf(stderr, "got stat\n");  */
  if (stat(dest_val,&dbuf) 
   || sbuf.st_size != dbuf.st_size
   || diff(sour_val,dest_val)) { 
    char newname[256]; 
    char ch; 
/* fprintf(stderr, "not a match\n");  */
    strcpy(newname, dest_val); 
    strcat(newname, "-tmp"); 
/* fprintf(stderr, "creating %s\n", newname);  */
    if (! (sfd=fopen(sour_val,"r"))) PL_fail; 
/* fprintf(stderr, "can read source\n");  */
    if (! (dfd=fopen(newname,"w"))) { fclose(sfd); PL_fail; } 
/* fprintf(stderr, "can write dest\n");  */
    /* yes, I know this is dumb */ 
    while (!feof(sfd) && (ch=getc(sfd))!= EOF) putc(ch,dfd); 
    fclose(sfd); 
    fclose(dfd); 
/* fprintf(stderr, "finished copy\n");  */
    if (chmod(newname,sbuf.st_mode)) PL_fail; 
/* fprintf(stderr, "finished chmod\n");  */
    times.actime=sbuf.st_atime; 
    times.modtime=sbuf.st_mtime; 
    if (utime(newname, &times)) PL_fail; 
/* fprintf(stderr, "finished utime\n");  */
    chown(newname,sbuf.st_uid,sbuf.st_gid); 
/* fprintf(stderr, "finished chown\n");  */
    if (rename(newname,dest_val)) PL_fail; 
/* fprintf(stderr, "finished rename\n");  */
    PL_succeed; 
  } else { 
/* fprintf(stderr, "does exist\n");  */
    if (sbuf.st_mtime != dbuf.st_mtime) { 
/* fprintf(stderr, "doing utime\n");  */
      times.actime=sbuf.st_atime; 
      times.modtime=sbuf.st_mtime; 
      if (utime(dest_val, &times)) PL_fail; 
/* fprintf(stderr, "finished utime\n");  */
    } 
    if (sbuf.st_mode != dbuf.st_mode) { 
/* fprintf(stderr, "doing chmod\n");  */
      if (chmod(dest_val,sbuf.st_mode)) PL_fail; 
/* fprintf(stderr, "finished chmod\n");  */
    } 
    if (sbuf.st_uid != dbuf.st_uid
     || sbuf.st_gid != dbuf.st_gid) { 
/* fprintf(stderr, "doing chown\n");  */
      chown(dest_val,sbuf.st_uid,sbuf.st_gid); 
/* fprintf(stderr, "finished chown\n");  */
    } 
    PL_succeed; 
  } 
  PL_succeed; 
} 

#define is_file(mode) (((mode)&S_IFMT)==S_IFREG)
#define is_dir(mode) (((mode)&S_IFMT)==S_IFDIR)
#define is_link(mode) (((mode)&S_IFMT)==S_IFLNK)

/* make a symbolic link */ 
foreign_t pl_link(term_t node, term_t target) {
  char *node_val; char *target_val; struct stat nbuf, lbuf; 
  if (!PL_get_atom_chars(node,&node_val)) PL_fail; 
  if (!PL_get_atom_chars(target,&target_val)) PL_fail; 
  if (stat(node_val,&nbuf)) PL_fail;
  if (!lstat(target_val, &lbuf)) { /* target exists */ 
/* fprintf(stderr, "target exists\n");  */
    if (is_link(lbuf.st_mode)) { 
      char buf[256]; 
/* fprintf(stderr, "is a link\n");  */
buf[0]=0; 
      if (readlink(target_val, buf, 256)<1) PL_fail; 
/* fprintf(stderr, "got buf %s\n", buf);  */
/* fprintf(stderr, "read link\n");  */
      if (strcmp(buf,node_val)!=0) { 
/* fprintf(stderr, "link different\n");  */
	if (unlink(target_val)) PL_fail; 
/* fprintf(stderr, "unlinked target\n");  */
        if (symlink(node_val, target_val)) PL_fail; 
/* fprintf(stderr, "linked new source\n");  */
	PL_succeed; 
      } else PL_succeed; 
    } else PL_fail; 		/* don't squash non-symlinks */ 
  } else { 			/* no such target */ 
    if (symlink(node_val, target_val)) PL_fail; 
  } 
  PL_succeed; 
}

install_t
pl_copy_install()
{ 
  PL_register_foreign("link", 2, pl_link, 0);
  PL_register_foreign("copy", 2, pl_copy, 0);
}
