/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
/* #include <signal.h> */ 
#include <sys/wait.h> 

foreign_t 
pl_email(term_t target, term_t message, term_t subject) 
{ 
  int pid; 
  int stat; 
  int filedes[2]; 
  char *target_val; 
  char *message_val; 
  char *subject_val; 

  if (!PL_get_atom_chars(target,&target_val)) PL_fail; 
  if (!PL_get_atom_chars(message,&message_val)) PL_fail; 
  if (!PL_get_atom_chars(subject,&subject_val)) PL_fail; 

  if (pipe(filedes)) PL_fail; 

  if ((pid=fork())) { 	/* parent */ 
    write(filedes[1],message_val,strlen(message_val)); 
    write(filedes[1],"\n",1); 
    close(filedes[1]); 
    close(filedes[0]); 
    /* close(filedes[0]); */
    waitpid(pid,&stat,0);
    if (stat==0) PL_succeed; 
    else PL_fail; 
  } else { 	  	/* child */ 
    char *args[128]; 
    char buf[1024], *p, *q; 
    int arg = 3; 
    int end = 0; 
    close(filedes[1]); 	/* arrange for close of output to close pipe */ 
    close(0); 		/* standard input */ 
    if (dup(filedes[0])!=0) { 
      fprintf(stderr,"close followed by dup failed!\n"); 
      exit(1); 
    } 
    /* construct argument list */ 
    args[0]="/usr/bin/mailx"; 
    args[1]="-s"; 
    args[2]=subject_val; 
    strcpy(buf,target_val); 
    p=buf; 
    while (!end) { 
      while (*p==' ' || *p=='\t' || *p=='\n') p++; 
      q=p; 
      while (*q!=' ' && *q!='\t' && *q!='\n' && *q!='\0') q++; 
      if (!*q) end=1; 
      *q=0; 
      args[arg++]=p; 
      p=q+1; 
    } 
    args[arg++]=NULL; 
    execv("/usr/bin/mailx", args); 
    fprintf(stderr,"exec failed!\n");   
    exit(1); 
  } 
  PL_succeed; 
} 

install_t
pl_email_install()
{ 
  PL_register_foreign("email", 3, pl_email, 0);
}
