/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h> 
#include <sys/wait.h> 

struct li_listelem {
  struct li_listelem *next;
  char *line; 
} ;

struct li_context {
  struct li_listelem *head, *tail;
} ;

/* This program executes a command and 
   returns its output one line at a time */ 

foreign_t 
pl_file_head(term_t file, term_t show, term_t output, foreign_t handle) 
{ 
  char *file_val; 
  int show_val; 
  struct li_context *ctxt; 
  FILE *in; 

  /* get parameters */ 
  if (!PL_get_atom_chars(file,&file_val)) PL_fail; 
  if (!PL_get_integer(show,&show_val)) PL_fail; 

  /* if this is the first call, then read command output */ 
  switch( PL_foreign_control(handle) )
  { case PL_FIRST_CALL:

      /* open file, read, close */ 
      ctxt = (struct li_context *) malloc(sizeof(struct li_context)); 
      ctxt->head = ctxt->tail = NULL; 
      in = fopen(file_val, "r"); 
      if (in) { 
        int count=0; 
	while (!feof(in) && count<show_val) { 
	  char buf[1024]; 
	  int len; 
 	  struct li_listelem *l; 
	  if (fgets(buf,1024,in)) { 
	    len=strlen(buf); 
	    if (buf[len-1]=='\n') buf[len-1]='\0'; 
	    l = (struct li_listelem *) malloc(sizeof(struct li_listelem)); 
	    l->line = (char *) malloc (strlen(buf)+1); 
	    strcpy(l->line,buf); 
	    if (!ctxt->head) ctxt->head=ctxt->tail=l;
	    else { ctxt->tail->next = l;  ctxt->tail=l; }
	    l->next = NULL;
          } 
	} 
      } 
      /* close input side of pipe */ 
      fclose(in); 

      /* now try to unify with requested lines */ 
      while (ctxt->head) {
	fid_t frame = PL_open_foreign_frame();
	if (PL_unify_atom_chars(output,ctxt->head->line)) {
	   struct li_listelem *p = ctxt->head;
	   ctxt->head = ctxt->head->next;
	   free(p->line); free(p); 
	   PL_close_foreign_frame(frame);
	   PL_retry_address(ctxt);
	}  else {
	   struct li_listelem *p = ctxt->head;
	   ctxt->head = ctxt->head->next;
	   free(p->line); free(p); 
	   PL_discard_foreign_frame(frame);
	}
      }
      PL_fail; 
      break; 
    case PL_REDO:
      /* next element is always the first element of the list */
      ctxt = PL_foreign_context_address(handle);
      while (ctxt->head) {
	fid_t frame = PL_open_foreign_frame();
	if (PL_unify_atom_chars(output,ctxt->head->line)) {
	   struct li_listelem *p = ctxt->head;
	   ctxt->head = ctxt->head->next;
	   free(p->line); free(p); 
	   PL_close_foreign_frame(frame);
	   PL_retry_address(ctxt);
	}  else {
	   struct li_listelem *p = ctxt->head;
	   ctxt->head = ctxt->head->next;
	   free(p->line); free(p); 
	   PL_discard_foreign_frame(frame);
	}
      }
      PL_fail; 
      break; 
    case PL_CUTTED: 
      while (ctxt->head) { 
        struct li_listelem *p = ctxt->head;
        ctxt->head = ctxt->head->next;
        free(p->line); free(p); 
      } 
      PL_fail; 
      break; 
  } 
  PL_succeed; 
} 

install_t
pl_file_head_install()
{ 
  PL_register_foreign("file_head", 3, pl_file_head, PL_FA_NONDETERMINISTIC);
}
