/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h> 

/* change the mode of a file to match a template, if necessary */ 
foreign_t 
pl_path_owner(term_t path, term_t uid, term_t gid) 
{ 
  char *path_val; 
  int uid_val, gid_val; 
  int got_uid, got_gid; 
  struct stat buf; 
  if (!PL_get_atom_chars(path,&path_val)) PL_fail; 
  if (stat(path_val,&buf)!=0) PL_fail; 
  got_uid = PL_get_integer(uid,&uid_val); 
  got_gid = PL_get_integer(gid,&gid_val); 
  if (!got_uid) { 
    uid_val=-1; /* don't change */ 
    if (!PL_unify_integer(uid,buf.st_uid)) PL_fail; 
  } 
  if (!got_gid) { 
    gid_val=-1; /* don't change */ 
    if (!PL_unify_integer(gid,buf.st_gid)) PL_fail; 
  } 
  if (((uid_val != -1) && (uid_val != buf.st_uid))
   || ((gid_val != -1) && (gid_val != buf.st_gid))) {
    if (!chown(path_val, uid_val, gid_val)) PL_succeed; 
    else PL_fail; 
  } else PL_succeed; 
} 

/* same thing for symbolic links */ 

foreign_t 
pl_path_lowner(term_t path, term_t uid, term_t gid) 
{ 
  char *path_val; 
  int uid_val, gid_val; 
  int got_uid, got_gid; 
  struct stat buf; 
  if (!PL_get_atom_chars(path,&path_val)) PL_fail; 
  if (lstat(path_val,&buf)!=0) PL_fail; 
  got_uid = PL_get_integer(uid,&uid_val); 
  got_gid = PL_get_integer(gid,&gid_val); 
  if (!got_uid) { 
    uid_val=-1; /* don't change */ 
    if (!PL_unify_integer(uid,buf.st_uid)) PL_fail; 
  } 
  if (!got_gid) { 
    gid_val=-1; /* don't change */ 
    if (!PL_unify_integer(gid,buf.st_gid)) PL_fail; 
  } 
  if (((uid_val != -1) && (uid_val != buf.st_uid))
   || ((gid_val != -1) && (gid_val != buf.st_gid))) {
    if (!lchown(path_val, uid_val, gid_val)) PL_succeed; 
    else PL_fail; 
  } else PL_succeed; 
} 

install_t
pl_owner_install()
{ 
  PL_register_foreign("path_owner", 3, pl_path_owner, 0);
  PL_register_foreign("path_lowner", 3, pl_path_lowner, 0);
}

