/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <grp.h> 

/* general idea is that I wish to delay unification for
   a set of variables unless all match.  So I'm building lists and
   unifying the lists.  This FAILS if one of the variables in the top
   level query is '_'.  PL_unify goes into an infinite loop during the
   redo (!) */

struct test_context { 
  int which; 
} ; 

foreign_t
pl_test(
      term_t t1,
      term_t t2, 
      foreign_t handle)
{
  term_t vars = PL_new_term_ref(); 
  term_t list = PL_new_term_ref(); 
  term_t elem = PL_new_term_ref(); 
  struct test_context *ctxt; 
  switch( PL_foreign_control(handle) )
  { case PL_FIRST_CALL:

      ctxt = (struct test_context *) malloc (sizeof(struct test_context)); 
      ctxt->which=1; 

      /* make a list of the variables to bind : build IN REVERSE */ 
      PL_put_nil(vars); 
      PL_cons_list(vars,t2,vars); 
      PL_cons_list(vars,t1,vars); 

      /* desired effect is delayed unification ONLY IF ALL MATCH. 
         So unify LISTS instead of atoms. */ 

      while(ctxt->which<6) { 
	PL_put_nil(list); 
	PL_put_integer(elem,ctxt->which+1); PL_cons_list(list,elem,list); 
	PL_put_integer(elem,ctxt->which);   PL_cons_list(list,elem,list); 
	ctxt->which++; 
	if (PL_unify(vars,list)) PL_retry_address(ctxt);
      } 
      free(ctxt); 
      PL_fail; 

    case PL_REDO:

      ctxt = PL_foreign_context_address(handle); 

      /* make a list of the variables to bind : build IN REVERSE */ 
      PL_put_nil(vars); 
      PL_cons_list(vars,t2,vars); 
      PL_cons_list(vars,t1,vars); 

      while (ctxt->which<6) { 

	PL_put_nil(list); 
	PL_put_integer(elem,ctxt->which+1); PL_cons_list(list,elem,list); 
	PL_put_integer(elem,ctxt->which);   PL_cons_list(list,elem,list); 
	ctxt->which++; 
	if (PL_unify(vars,list)) PL_retry_address(ctxt);
      } 
      free(ctxt); 
      PL_fail; 

    case PL_CUTTED:

      /* cut context */ 
      ctxt = PL_foreign_context_address(handle); 
      free(ctxt); 
      PL_succeed;
  }
  PL_fail; 
} 

install_t
pl_test_install()
{ 
  PL_register_foreign("test", 2, pl_test, PL_FA_NONDETERMINISTIC);
}
